/*
 *  Copyright 2001-2004 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#define YY_NEVER_INTERACTIVE 1
//#define WANT_TOKEN_WRITE

#include <iostream>
#include "ragel.h"
#include "ptreetypes.h"
#include "rlparse.h"
#include "buffer.h"

using std::cout;
using std::cerr;
using std::endl;

Buffer tokbuf;
int builtinBrace = 0;
bool inlineWhitespace = true;

int first_line = 1;
int first_column = 1;
int last_line = 1;
int last_column = 0;

/* These keep track of the start of an inline comment or literal string for
 * reporting unterminated comments or strings. */
int il_comm_lit_first_line;
int il_comm_lit_first_column;

/* These keep track of the start of a code block for reporting unterminated
 * code blocks. */
int il_code_first_line;
int il_code_first_column;

extern YYLTYPE yylloc;

void extendGarble();
void garble();

void extendToken( char *data, int len );
int emitToken( int token, char *data, int len );
int emitNoData( int token );
void passThrough( char *data );

enum InlineBlockType {
	CurlyDelimited,
	SemiTerminated
} inlineBlockType;

%}

/* Outside an fsm machine specification ("outside code"). */
%x		OC_SGL_LIT
%x		OC_DBL_LIT
%x		OC_C_COM
%x		OC_CXX_COM

/* Inside a fsm machine specification. */
%x		RL_INITIAL
%x		RL_SLIT
%x		RL_DLIT
%x		RL_OREXP
%x		RL_REGEXP
%x		RL_REGEXP_OR
%x		RL_SHELL_COM
%x		RL_BASE_CLAUSE

/* Inline code. */
%x		IL_INITIAL
%x		IL_SGL_LIT
%x		IL_DBL_LIT
%x		IL_C_COM
%x		IL_CXX_COM

WHITESPACE [\t\n\v\f\r ]

%%

	/* Numbers in outter code. */
<INITIAL>[0-9]+ {
	garble();
	passThrough( yytext );
}
	/* Words in outter code. */
<INITIAL>[A-Za-z_][A-Za-z_0-9]* {
	garble();
	passThrough( yytext );
}

	/* Begin a c style comment. */
<INITIAL>"/*" {
	BEGIN(OC_C_COM);
	extendGarble();
	passThrough( yytext );
}
	/* Data in a C style comment. */
<OC_C_COM>. 		extendGarble(); passThrough( yytext );
<OC_C_COM>\n 		extendGarble(); passThrough( yytext );

	/* Terminate a C style comment. */
<OC_C_COM>"*/" {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}

	/* Begin a C++ style comment. */
<INITIAL>"//" {
	BEGIN(OC_CXX_COM);
	extendGarble();
	passThrough( yytext );
}
	/* Data in a C++ style comment. */
<OC_CXX_COM>[^\n]+ {
	extendGarble();
	passThrough( yytext );
}
	/* Terminate a C++ style comment. */
<OC_CXX_COM>\n {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}


	/* Start literals. */
<INITIAL>\' {
	BEGIN(OC_SGL_LIT);
	extendGarble();
	passThrough( yytext );
}
<INITIAL>\" {
	BEGIN(OC_DBL_LIT);
	extendGarble();
	passThrough( yytext );
}
	/* Various escape sequences in literals. We don't need to get them
	 * all here. We just need to pick off the ones that could confuse us
	 * about the literal we are matchine */
<OC_SGL_LIT,OC_DBL_LIT>\\\'		extendGarble(); passThrough( yytext );
<OC_SGL_LIT,OC_DBL_LIT>\\\"		extendGarble(); passThrough( yytext );
<OC_SGL_LIT,OC_DBL_LIT>\\\\		extendGarble(); passThrough( yytext );
	/* Characters in literals. */
<OC_DBL_LIT>[^\"]				extendGarble(); passThrough( yytext );
<OC_SGL_LIT>[^\']				extendGarble(); passThrough( yytext );
	/* Terminate a double literal */
<OC_DBL_LIT>\" {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}
	/* Terminate a single literal. */
<OC_SGL_LIT>\' {
	BEGIN(INITIAL);
	garble();
	passThrough( yytext );
}

	/* Whitespace. */
<INITIAL>{WHITESPACE}+ {
	garble();
	passThrough( yytext );
}

	/* Section Deliminator */
<INITIAL>"%%" {
	BEGIN(RL_INITIAL);
	return emitNoData( TK_Section );
}

	/* Any other characters. */
<INITIAL>. {
	garble();
	passThrough( yytext );
}

	/* Numbers. */
<RL_INITIAL,IL_INITIAL>[0-9][0-9]* {	
	return emitToken( TK_UInt, yytext, yyleng );
}
<RL_INITIAL,IL_INITIAL>\+[0-9][0-9]* {	
	return emitToken( TK_PInt, yytext, yyleng );
}
<RL_INITIAL,IL_INITIAL>-[0-9][0-9]* {	
	return emitToken( TK_NInt, yytext, yyleng );
}
<RL_INITIAL,IL_INITIAL>0x[0-9a-fA-F][0-9a-fA-F]* {	
	return emitToken( TK_Hex, yytext, yyleng );
}

	/* Keywords in RL and IL. */
<RL_INITIAL>init {
	return emitNoData( KW_Init );
}
<RL_INITIAL>action {
	return emitNoData( KW_Action );
}
<RL_INITIAL>struct {
	BEGIN(RL_BASE_CLAUSE);
	return emitNoData( KW_Struct );
}
<RL_INITIAL>alphtype {
	BEGIN(IL_INITIAL);
	inlineWhitespace = false;
	inlineBlockType = SemiTerminated;
	return emitNoData( KW_AlphType );
}
<RL_INITIAL>element {
	BEGIN(IL_INITIAL);
	inlineBlockType = SemiTerminated;
	return emitNoData( KW_Element );
}
<RL_INITIAL>getkey {
	BEGIN(IL_INITIAL);
	inlineBlockType = SemiTerminated;
	return emitNoData( KW_GetKey );
}
<RL_INITIAL>context {
	return emitNoData( KW_Context );
}
	/*<RL_INITIAL>range {
	return emitNoData( KW_Range );
} */

	/* These must be synced in rlparse.y */
<IL_INITIAL>fpc {
	return emitNoData( KW_PChar );
}
<IL_INITIAL>fc {
	return emitNoData( KW_Char );
}
<IL_INITIAL>fhold {
	return emitNoData( KW_Hold );
}
<IL_INITIAL>fgoto {
	return emitNoData( KW_Goto );
}
<IL_INITIAL>fcall {
	return emitNoData( KW_Call );
}
<IL_INITIAL>fret {
	return emitNoData( KW_Ret );
}
<IL_INITIAL>fstack {
	return emitNoData( KW_Stack );
}
<IL_INITIAL>fcurs {
	return emitNoData( KW_CurState );
}
<IL_INITIAL>ftargs {
	return emitNoData( KW_TargState );
}
<IL_INITIAL>fentry {
	return emitNoData( KW_Entry );
}
<IL_INITIAL>fnext {
	return emitNoData( KW_Next );
}
<IL_INITIAL>fexec {
	return emitNoData( KW_Exec );
}
<IL_INITIAL>fbuf {
	return emitNoData( KW_Buf );
}
<IL_INITIAL>fblen {
	return emitNoData( KW_BLen );
}

	/* Words. */
<RL_INITIAL,IL_INITIAL>[A-Za-z_][A-Za-z_0-9]* {
	return emitToken( TK_Word, yytext, yyleng );
}

	/* Begin a shell style comment. */
<RL_INITIAL>#			{
	BEGIN(RL_SHELL_COM);
	extendGarble();
}
	/* Data in a shell style comment. */
<RL_SHELL_COM>[^\n]+		{
	extendGarble();
}
	/* Terminate a C++ style comment. */
<RL_SHELL_COM>\n		{
	BEGIN(RL_INITIAL);
	garble();
}

	/* Content of Base Clause. If there is only whitespace up until the first {
	 * then it will be garbled. If there is anything else then the whole thing
	 * will all be sent as TK_BaseClause. */
<RL_BASE_CLAUSE>{WHITESPACE}+ garble();
<RL_BASE_CLAUSE>[^{]+ {
	return emitToken( TK_BaseClause, yytext, yyleng );
}

	/* 
	 * Start single and double literals.
	 */
<RL_INITIAL>'			{
	BEGIN(RL_SLIT);
	extendToken( 0, 0 );
}
<RL_INITIAL>\"			{
	BEGIN(RL_DLIT);
	extendToken( 0, 0 );
}

	/* Escape sequences in single and double literals. */
<RL_SLIT,RL_DLIT>\\0		extendToken( "\0", 1 );
<RL_SLIT,RL_DLIT>\\a		extendToken( "\a", 1 );
<RL_SLIT,RL_DLIT>\\b		extendToken( "\b", 1 );
<RL_SLIT,RL_DLIT>\\t		extendToken( "\t", 1 );
<RL_SLIT,RL_DLIT>\\n		extendToken( "\n", 1 );
<RL_SLIT,RL_DLIT>\\v		extendToken( "\v", 1 );
<RL_SLIT,RL_DLIT>\\f		extendToken( "\f", 1 );
<RL_SLIT,RL_DLIT>\\r		extendToken( "\r", 1 );
<RL_SLIT,RL_DLIT>\\\n		extendToken( 0, 0 );
<RL_SLIT,RL_DLIT>\\.		extendToken( yytext+1, 1 );

	/* Characters in literals. */
<RL_SLIT>[^']						extendToken( yytext, 1 );
<RL_DLIT>[^"]						extendToken( yytext, 1 );

	/* Terminate a single literal. */
<RL_SLIT>' {
	BEGIN(RL_INITIAL);
	return emitToken( TK_Literal, 0, 0 );
}
	/* Terminate a double literal */
<RL_DLIT>\" {
	BEGIN(RL_INITIAL);
	return emitToken( TK_Literal, 0, 0 );
}

	/*
	 * Start an OR expression. 
	 */
<RL_INITIAL>"["			{
	BEGIN(RL_OREXP);
	return emitNoData( RE_SqOpen );
}

<RL_INITIAL>"\[^"	{
	BEGIN(RL_OREXP);
	return emitNoData( RE_SqOpenNeg );
}

	/* Escape sequences in OR expressions. */
<RL_OREXP>\\0		{ return emitToken( RE_Char, "\0", 1 ); }
<RL_OREXP>\\a		{ return emitToken( RE_Char, "\a", 1 ); }
<RL_OREXP>\\b		{ return emitToken( RE_Char, "\b", 1 ); }
<RL_OREXP>\\t		{ return emitToken( RE_Char, "\t", 1 ); }
<RL_OREXP>\\n		{ return emitToken( RE_Char, "\n", 1 ); }
<RL_OREXP>\\v		{ return emitToken( RE_Char, "\v", 1 ); }
<RL_OREXP>\\f		{ return emitToken( RE_Char, "\f", 1 ); }
<RL_OREXP>\\r		{ return emitToken( RE_Char, "\r", 1 ); }
<RL_OREXP>\\\n		{ garble(); }
<RL_OREXP>\\.		{ return emitToken( RE_Char, yytext+1, 1 ); }

	/* Range dash in an OR expression. */
<RL_OREXP>-	{
	return emitNoData( RE_Dash );
}

	/* Characters in an OR expression. */
<RL_OREXP>[^\]] {
	return emitToken( RE_Char, yytext, 1 );
}

	/* Terminate an OR expression. */
<RL_OREXP>\]	{
	BEGIN(RL_INITIAL); 
	return emitNoData( RE_SqClose );
}

	/* 
	 * Start a regular expression. 
	 */
<RL_INITIAL>\/		{
	BEGIN(RL_REGEXP);
	return emitNoData( RE_Slash );
}

	/* Escape sequences in regular expressions. */
<RL_REGEXP,RL_REGEXP_OR>\\0		{
	return emitToken( RE_Char, "\0", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\a		{
	return emitToken( RE_Char, "\a", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\b		{
	return emitToken( RE_Char, "\b", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\t		{
	return emitToken( RE_Char, "\t", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\n		{
	return emitToken( RE_Char, "\n", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\v		{
	return emitToken( RE_Char, "\v", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\f		{
	return emitToken( RE_Char, "\f", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\r		{
	return emitToken( RE_Char, "\r", 1 );
}
<RL_REGEXP,RL_REGEXP_OR>\\\n	{
	garble();
}
<RL_REGEXP,RL_REGEXP_OR>\\.		{
	return emitToken( RE_Char, yytext+1, 1 );
}

	/* Special characters in a regular expression. */
<RL_REGEXP>\.		{
	return emitNoData( RE_Dot );
}
<RL_REGEXP>\*		{
	return emitNoData( RE_Star );
}
<RL_REGEXP>"\[^"	{
	BEGIN(RL_REGEXP_OR);
	return emitNoData( RE_SqOpenNeg );
}
<RL_REGEXP>"\["		{
	BEGIN(RL_REGEXP_OR);
	return emitNoData( RE_SqOpen );
}

	/* Range dash in a regular expression or set. */
<RL_REGEXP_OR>-	{
	return emitNoData( RE_Dash );
}

	/* Terminate an or set or a regular expression. */
<RL_REGEXP_OR>\]	{
	BEGIN(RL_REGEXP); 
	return emitNoData( RE_SqClose );
}

	/* Characters in a regular expression. */
<RL_REGEXP,RL_REGEXP_OR>[^/]			{
	return emitToken( RE_Char, yytext, 1 );
}

	/* Terminate a regular expression */
<RL_REGEXP,RL_REGEXP_OR>\/ {
	BEGIN(RL_INITIAL);
	return emitNoData( RE_Slash );
}

	/*
	 * Whitespace. 
	 */
<RL_INITIAL>{WHITESPACE}+		garble();
	
	/* Builtin code move to Builtin initial. */
<RL_INITIAL,RL_BASE_CLAUSE>"{" {
	/* Start an inline code block. Keep track of where it started in
	 * case it terminates prematurely. Return the open bracket. */
	BEGIN(IL_INITIAL);
	inlineBlockType = CurlyDelimited;
	il_code_first_line = last_line;
	il_code_first_column = last_column+1;
	builtinBrace++;
	return emitNoData( *yytext );
}

	/* The double dot, we interpret this in the scanner becuase an
	 * LALR(1) parser cannot handle it, need LR(2). */
<RL_INITIAL>\.\. {
	return emitNoData( TK_DoubleDot );
}

	/* The instantiation operator. */
<RL_INITIAL>:= {
	return emitNoData( TK_ColonEquals );
}

	/* Error actions. */
<RL_INITIAL>\>\! {
	return emitNoData( TK_StartGblError );
}
<RL_INITIAL>@\! {
	return emitNoData( TK_FinishGblError );
}
<RL_INITIAL>\$\! {
	return emitNoData( TK_AllGblError );
}
<RL_INITIAL>%\! {
	return emitNoData( TK_LeaveGblError );
}
<RL_INITIAL>\>@\! {
	return emitNoData( TK_StartFinishGblError );
}
<RL_INITIAL>@%\! {
	return emitNoData( TK_FinishLeaveGblError );
}

	/* Local error actions. */
<RL_INITIAL>\>\^ {
	return emitNoData( TK_StartLocalError );
}
<RL_INITIAL>@\^ {
	return emitNoData( TK_FinishLocalError );
}
<RL_INITIAL>\$\^ {
	return emitNoData( TK_AllLocalError );
}
<RL_INITIAL>%\^ {
	return emitNoData( TK_LeaveLocalError );
}
<RL_INITIAL>\>@\^ {
	return emitNoData( TK_StartFinishLocalError );
}
<RL_INITIAL>@%\^ {
	return emitNoData( TK_FinishLeaveLocalError );
}

	/* Context embedding. */
<RL_INITIAL>\>: {
	return emitNoData( TK_StartContext );
}
<RL_INITIAL>@: {
	return emitNoData( TK_FinishContext );
}
<RL_INITIAL>\$: {
	return emitNoData( TK_AllContext );
}
<RL_INITIAL>%: {
	return emitNoData( TK_LeaveContext );
}
<RL_INITIAL>\>@: {
	return emitNoData( TK_StartFinishContext );
}
<RL_INITIAL>@%: {
	return emitNoData( TK_FinishLeaveContext );
}


	/* The arrow operator. */
<RL_INITIAL>-> {
	return emitNoData( TK_Arrow );
}

	/* Double star (longest match kleene star). */
<RL_INITIAL>\*\* {
	return emitNoData( TK_StarStar );
}

<RL_INITIAL>:: {
	return emitNoData( TK_NameSep );
}

	/* Any other characters. */
<RL_INITIAL>. {
	return emitNoData( *yytext );
}

	/* Section Deliminator */
<RL_INITIAL>"%%" {
	BEGIN(INITIAL);
	return emitNoData( TK_Section );
}

	/* End of input in a literal is an error. */
<RL_SLIT,RL_DLIT><<EOF>> {
	error(first_line, first_column) << "unterminated literal" << endl;
	exit(1);
}
	
	/* End of input in a comment is an error. */
<RL_SHELL_COM><<EOF>> {
	error(first_line, first_column) << "unterminated comment" << endl;
	exit(1);
}

	/* Begin a C style comment. */
<IL_INITIAL>"/*" {
	BEGIN(IL_C_COM);
	il_comm_lit_first_line = last_line;
	il_comm_lit_first_column = last_column+1;
	extendToken( yytext, yyleng );
}
	/* Data in a C style comment. */
<IL_C_COM>\n 	extendToken( yytext, 1 );
<IL_C_COM>. 	extendToken( yytext, 1 );

	/* Terminate a C style comment. */
<IL_C_COM>"*/" {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Comment, yytext, 2 );
}

	/* Begin a C++ style comment. */
<IL_INITIAL>"//" {
	BEGIN(IL_CXX_COM);
	il_comm_lit_first_line = last_line;
	il_comm_lit_first_column = last_column+1;
	extendToken( yytext, yyleng );
}
	/* Data in a C++ style comment. */
<IL_CXX_COM>[^\n]+ {
	extendToken( yytext, yyleng );
}
	/* Terminate a C++ style comment. */
<IL_CXX_COM>\n {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Comment, yytext, 1 );
}


	/* Start literals. */
<IL_INITIAL>' {
	BEGIN(IL_SGL_LIT);
	il_comm_lit_first_line = last_line;
	il_comm_lit_first_column = last_column+1;
	extendToken( yytext, 1 );
}
<IL_INITIAL>\" {
	BEGIN(IL_DBL_LIT);
	il_comm_lit_first_line = last_line;
	il_comm_lit_first_column = last_column+1;
	extendToken( yytext, 1 );
}
	/* Various escape sequences in literals. We don't need to get them
	 * all here. We just need to pick off the ones that could confuse us
	 * about the literal we are matching */
<IL_SGL_LIT,IL_DBL_LIT>\\'		extendToken( yytext, yyleng );
<IL_SGL_LIT,IL_DBL_LIT>\\\"		extendToken( yytext, yyleng );
<IL_SGL_LIT,IL_DBL_LIT>\\\\		extendToken( yytext, yyleng );
	/* Characters in literals. */
<IL_DBL_LIT>[^\"]				extendToken( yytext, 1 );
<IL_SGL_LIT>[^']				extendToken( yytext, 1 );

	/* Terminate a double literal */
<IL_DBL_LIT>\" {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Literal, yytext, 1 );
}
	/* Terminate a single literal. */
<IL_SGL_LIT>' {
	BEGIN(IL_INITIAL);
	return emitToken( IL_Literal, yytext, 1 );
}

	/* Open Brace, increment count of open braces. */
<IL_INITIAL>"{" {
	builtinBrace++;
	return emitToken( IL_Symbol, yytext, 1 );
}

	/* Close brace, decrement count of open braces. */
<IL_INITIAL>"}" {
	builtinBrace--;
	if ( inlineBlockType == CurlyDelimited && builtinBrace == 0 ) {
		/* Inline code block ends. */
		BEGIN(RL_INITIAL);
		inlineWhitespace = true;
		return emitNoData( *yytext );
	}
	else {
		/* Either a semi terminated inline block or only the closing brace of
		 * some inner scope, not the block's closing brace. */
		return emitToken( IL_Symbol, yytext, 1 );
	}
}

	/* May need to terminate the inline block. */
<IL_INITIAL>; {
	if ( inlineBlockType == SemiTerminated ) {
		/* Inline code block ends. */
		BEGIN(RL_INITIAL);
		inlineWhitespace = true;
		return emitNoData( TK_Semi );
	}
	else {
		/* Not ending. The semi is sent as a token, not a generic symbol. */
		return emitNoData( *yytext );
	}
}

	/* Catch some symbols so they can be 
	 * sent as tokens instead as generic symbols. */
<IL_INITIAL>[\[\],()\*] {
	return emitNoData( *yytext );
}
<IL_INITIAL>:: {
	return emitNoData( TK_NameSep );
}

	/* Whitespace. */
<IL_INITIAL>{WHITESPACE}+ {
	if ( inlineWhitespace )
		return emitToken( IL_WhiteSpace, yytext, yyleng );
}

	/* Any other characters. */
<IL_INITIAL>. {
	return emitToken( IL_Symbol, yytext, 1 );
}


	/* End of input in a literal is an error. */
<IL_SGL_LIT,IL_DBL_LIT><<EOF>>		{
	error(il_comm_lit_first_line, il_comm_lit_first_column) << 
			"unterminated literal" << endl;
	exit(1);
}
	
	/* End of input in a comment is an error. */
<IL_C_COM,IL_CXX_COM><<EOF>>	{
	error(il_comm_lit_first_line, il_comm_lit_first_column) <<
			"unterminated comment" << endl;
	exit(1);
}

	/* End of intput in a code block. */
<IL_INITIAL><<EOF>> {
	error(il_code_first_line, il_code_first_column) <<
			"unterminated code block" << endl;
	exit(1);
}

%%

/* Write out token data, escaping special charachters. */
#ifdef WANT_TOKEN_WRITE
void writeToken( int token, char *data )
{
	cout << "token id " << token << " at " << 
			yylloc.first_line << ":" << yylloc.first_column << "-" <<
			yylloc.last_line << ":" << yylloc.last_column << " ";

	if ( data != 0 ) {
		while ( *data != 0 ) {
			switch ( *data ) {
			case '\n':	cout << "\\n"; break;
			case '\t':	cout << "\\t"; break;
			default:	cout << *data; break;
			}
			data += 1;
		}
	}
	cout << endl;
}
#endif

/* Caclulate line info from yytext. Called on every pattern match. */
void updateLineInfo()
{
	/* yytext should always have at least wone char. */
	assert( yytext[0] != 0 );

	/* Scan through yytext up to the last character. */
	char *p = yytext;
	for ( ; p[1] != 0; p++ ) {
		if ( p[0] == '\n' ) {
			last_line += 1;
			last_column = 0;
		}
		else {
			last_column += 1;
		}
	}

	/* Always consider the last character as not a newline. Newlines at the
	 * end of a token are as any old character at the end of the line. */
	last_column += 1;

	/* Save off the line information because the caller may emit a token. */
	yylloc.first_line = first_line;
	yylloc.first_column = first_column;
	yylloc.last_line = last_line;
	yylloc.last_column = last_column;

	/* If the last character was indeed a newline, then wrap ahead now. */
	if ( p[0] == '\n' ) {
		last_line += 1;
		last_column = 0;
	}
}

/* Extend the garbling of a matched pattern that will not be part of a token.
 * Provided so that line info of garbled tokens can be preserved across
 * multiple patterns. */
void extendGarble() 
{
	/* Update line information from yytext. */
	updateLineInfo();
}

/* Eat up a matched pattern that will not be part of a token. */
void garble() 
{
	/* Update line information from yytext. */
	updateLineInfo();

	/* The next token starts ahead of the last token. */
	first_line = last_line;
	first_column = last_column + 1;
}

/* Append data to the end of the token. More token data expected. */
void extendToken( char *data, int len )
{
	if ( data != 0 && len > 0 )
		tokbuf.append( data, len );

	/* Update line information from yytext. */
	updateLineInfo();
}

/* Append data to the end of a token and emitToken it to the parser. */
int emitToken( int token, char *data, int len )
{
	/* Append the data and null terminate. */
	if ( data != 0 && len > 0 )
		tokbuf.append( data, len );
	tokbuf.append( 0 );

	/* Duplicate the buffer. */
	yylval.data = new char[tokbuf.length];
	strcpy( yylval.data, tokbuf.data );

	/* Update line information from yytext. */
	updateLineInfo();

	/* Write token info. */
#ifdef WANT_TOKEN_WRITE
	writeToken( token, tokbuf.data );
#endif

	/* Clear out the buffer. */
	tokbuf.clear();

	/* The next token starts ahead of the last token. */
	first_line = last_line;
	first_column = last_column + 1;

	return token;
}

/* Emit a token with no data to the parser. */
int emitNoData( int token ) 
{
	/* Return null to the parser. */
	yylval.data = 0;

	/* Update line information from yytext. */
	updateLineInfo();

	/* Write token info. */
#ifdef WANT_TOKEN_WRITE
	writeToken( token, 0 );
#endif

	/* Clear out the buffer. */
	tokbuf.clear();

	/* The next token starts ahead of the last token. */
	first_line = last_line;
	first_column = last_column + 1;

	return token;
}

/* Pass tokens in outter code through to the output. */
void passThrough( char *data )
{
	if ( gblErrorCount == 0 && (outputFormat == OutCCode || 
			outputFormat == OutCppCode || outputFormat == OutObjCCode) )
		*outStream << data;
}

/* Init a buffer. */
Buffer::Buffer() 
:
	data(0), 
	length(0),
	allocated(0)
{
}

/* Empty out a buffer on destruction. */
Buffer::~Buffer()
{
	empty();
}

/* Free the space allocated for the buffer. */
void Buffer::empty()
{
	if ( data != 0 ) {
		free( data );

		data = 0;
		length = 0;
		allocated = 0;
	}
}

/* Grow the buffer when to len allocation. */
void Buffer::upAllocate( int len )
{
	if ( data == 0 )
		data = (char*) malloc( len );
	else
		data = (char*) realloc( data, len );
	allocated = len;
}

int yywrap()
{
	/* Once processessing of the input is done, no more. */
	return 1;
}

/* Here simply to suppress the unused yyunpt warning. */
void thisFuncIsNeverCalled()
{
	yyunput(0, 0);
}

